/*
 * (C) Copyright 2000-2003
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * Misc boot support
 */
#include <common.h>
#include <command.h>
#include <asm/byteorder.h>
#include <net.h>


#define tag_offs(sz) ((struct tag *)((u32 *)params + sz))

static int setup_start_tag ( struct tag *params)
{
        params->hdr.tag = ATAG_CORE;
        params->hdr.size = tag_size (tag_core);

        params->u.core.flags = 0;
        params->u.core.pagesize = 0;
        params->u.core.rootdev = 0;

	return params->hdr.size;
}

static int setup_mem_tag ( struct tag *params)
{
	DECLARE_GLOBAL_DATA_PTR;
	int i = CONFIG_NR_DRAM_BANKS;
	int size = 0;
	
//	for(i=0; i<CONFIG_NR_DRAM_BANKS; i++)
	while( --i >= 0)
	{
		u32 bsize = gd->bd->bi_dram[i].size;

		if( bsize > 0)
		{
			params->hdr.tag = ATAG_MEM;
		        params->hdr.size = tag_size (tag_mem32);

			params->u.mem.start = gd->bd->bi_dram[i].start;
			params->u.mem.size = bsize;
		
			size += params->hdr.size;
			params = tag_offs (params->hdr.size);
		}
	}
	
	return size;
}

static int setup_commandline_tag ( char *commandline, struct tag *params)
{
        char *p;

        if (!commandline)
                return 0;

        /* eat leading white space */
        for (p = commandline; *p == ' '; p++);

        /* skip non-existent command lines so the kernel will still
         * use its default command line.
         */
        if (*p == '\0')
                return 0;

        params->hdr.tag = ATAG_CMDLINE;
        params->hdr.size =
                (sizeof (struct tag_header) + strlen (p) + 1 + 4) >> 2;

        strcpy (params->u.cmdline.cmdline, p);

	return params->hdr.size;
}

static int setup_end_tag ( struct tag *params)
{
        params->hdr.tag = ATAG_NONE;
        params->hdr.size = 0;

	return( sizeof(struct tag_header) >> 2);
}

/* -------------------------------------------------------------------- */

static int setup_tags( ulong addr)
{
    int size = 0;

    struct tag *params = (struct tag *) addr;

    size += setup_start_tag( params);
    size += setup_mem_tag( tag_offs(size) );
    size += setup_commandline_tag( getenv("bootargs"), tag_offs(size) );
    size += setup_end_tag( tag_offs(size) );

    return size;
}

int do_wtags (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	DECLARE_GLOBAL_DATA_PTR;

	bd_t *bd = gd->bd;
        ulong dest = bd->bi_boot_params;
 
        if (argc > 1)
                dest = simple_strtoul(argv[1], NULL, 16);

        // check if we are writing to flash
        if ( flash_addr2info(dest) != -1)
        {
                int rc = 0;
                ulong addr = bd->bi_boot_params;
                int size = setup_tags( addr);
         
                printf("Writing TAGS to FLASH at 0x%08lx\n", dest);
                
                if( (rc = flash_write ((uchar *) addr, dest, size << 2)) < 0)
                {
                        flash_perror(rc);
                        return 1;
                }
        }
        else
        {
                printf("Writing TAGS to DRAM at 0x%08lx\n", dest);
                setup_tags( dest);
        }

        return 0;
}

int do_go (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
#if defined(CONFIG_I386)
	DECLARE_GLOBAL_DATA_PTR;
#endif
	ulong	addr, rc;
	int     rcode = 0;

	if (argc < 2)
		addr = load_addr;
	else
		addr = simple_strtoul(argv[1], NULL, 16);

	printf ("## Starting application at 0x%08lX ...\n", addr);

	/*
	 * pass address parameter as argv[0] (aka command name),
	 * and all remaining args
	 */
#if defined(CONFIG_I386)
	/*
	 * x86 does not use a dedicated register to pass the pointer
	 * to the global_data
	 */
	argv[0] = (char *)gd;
#endif
#if !defined(CONFIG_NIOS)
	rc = ((ulong (*)(int, char *[]))addr) (--argc, &argv[1]);
#else
	/*
	 * Nios function pointers are address >> 1
	 */
	rc = ((ulong (*)(int, char *[]))(addr>>1)) (--argc, &argv[1]);
#endif
	if (rc != 0) rcode = 1;

	printf ("## Application terminated, rc = 0x%lX\n", rc);
	return rcode;
}

/* -------------------------------------------------------------------- */

U_BOOT_CMD(
	wtags, 2, 1,	do_wtags,
        "wtags   - write Linux kernel image tags\n",
        "[addr]\n    - write Linux kernel image tags at 'addr'\n"
);

U_BOOT_CMD(
	go, CFG_MAXARGS, 1,	do_go,
	"go      - start application at address 'addr'\n",
	"addr [arg ...]\n    - start application at address 'addr'\n"
	"      passing 'arg' as arguments\n"
);

extern int do_reset (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);

U_BOOT_CMD(
	reset, 1, 0,	do_reset,
	"reset   - Perform RESET of the CPU\n",
	NULL
);
