/*
 *  drivers/mtd/onenand/onenand_uboot.c
 *
 *  Copyright (C) 2005-2008 Samsung Electronics
 *  Kyungmin Park <kyungmin.park@samsung.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/*
 * OneNAND initialization at U-Boot
 */

#include <common.h>
#include <linux/mtd/compat.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/onenand.h>

#ifdef CFG_FLASH_ONENAND
struct mtd_info onenand_mtd;
struct onenand_chip onenand_chip;

int flash_addr2info (ulong addr)
{
	ulong base = (ulong) onenand_chip.base;
	return( addr >= base && addr <= base + onenand_chip.chipsize - 1)? 1:-1;
}

int flash_sector_count ( ulong bank)
{
	return (onenand_mtd.size >> onenand_chip.erase_shift);
}

unsigned long flash_init (void)
{
	memset(&onenand_mtd, 0, sizeof(struct mtd_info));
	memset(&onenand_chip, 0, sizeof(struct onenand_chip));

	onenand_chip.base = (void *) CFG_ONENAND_BASE;
	onenand_mtd.priv = &onenand_chip;

	puts("OneNAND: ");
	onenand_scan(&onenand_mtd, 1);
	
	return onenand_mtd.size;
}

void flash_print_info (int bank)
{
	uint i, sectors;

	printf("%s", onenand_mtd.name);
	printf ("\n  %s command set, Manufacturer ID: 0x%02X, Device ID: 0x%02X",
		(onenand_chip.manufact_id == 0xEC)? "SAMSUNG":"Unkonwn",
		onenand_chip.manufact_id, onenand_chip.device_id);
	printf ("\n  Erase block size: 0x%lX bytes, Page size: 0x%lX bytes\n",
		1 << onenand_chip.erase_shift, 1 << onenand_chip.page_shift);

	sectors = flash_sector_count(0);

        puts ("\n  Sector Start Addresses:");
        for (i = 0; i < sectors; ++i) {
                if ((i % 5) == 0)
                        printf ("\n");
		printf ("  %08lX      ", i << onenand_chip.erase_shift);
	}
	putc ('\n');
}

/*-----------------------------------------------------------------------
 *  * Copy memory to flash, returns:
 *   * 0 - OK
 *   * 1 - erase error
 */
int flash_erase (int bank, int s_first, int s_last)
{
	ulong block;
	struct erase_info instr = {
		.callback       = NULL,
	};

	if (s_last < 0)
        	s_last = flash_sector_count(bank) - 1;

	printf("\nErase block from %lu (%x) to %lu (%x)\n",
	        s_first, (s_first<<onenand_chip.erase_shift),
	        s_last, (((s_last+1)<<onenand_chip.erase_shift)-1) );

	for (block = s_first; block <= s_last; block++) {
		instr.addr = block << onenand_chip.erase_shift;
		instr.len = 1 << onenand_chip.erase_shift;
		if( onenand_erase(&onenand_mtd, &instr))
			printf(" erase failed at block %lu\n", block);
	}
	printf("Done\n");
	return 0;
}

/*-----------------------------------------------------------------------
 *  * Copy memory to flash, returns:
 *   * 0 - OK
 *   * 1 - write timeout
 *   * 2 - Flash not erased
 */
int flash_write (uchar *src, ulong addr, ulong cnt)
{
	size_t retlen = 0;

	onenand_write(&onenand_mtd, addr, cnt, &retlen,(u_char *) src);

	printf("Done\n");
	return 0;
}

int flash_read (uchar *dst, ulong addr, ulong cnt)
{
	struct mtd_oob_ops ops;

	ops.mode = MTD_OOB_PLACE;
	ops.len = cnt;
	ops.datbuf = (u_char *) dst;
	ops.ooblen = 0;
	ops.oobbuf = NULL;
	ops.retlen = ops.oobretlen = 0;

	onenand_mtd.read_oob(&onenand_mtd, addr, &ops);

	return 0;
}

#endif

